# Importing needed libraries

import os
import sys
import cv2  # OpenCV Library
import time
import torch
import numpy as np
from PIL import Image
from runpy import run_path
from math import log10, sqrt
import torch.nn.functional as F
from skimage import img_as_ubyte
from collections import OrderedDict
import torchvision.transforms.functional as TF


# Loading model and its weights and if prefered, set program to run on CPU
def load_checkpoint(model, weights):
    checkpoint = torch.load(weights, map_location='cpu')
    try:
        model.load_state_dict(checkpoint["state_dict"])
    except:
        state_dict = checkpoint["state_dict"]
        new_state_dict = OrderedDict()
        for k, v in state_dict.items():
            name = k[7:]  # remove `module.`
            new_state_dict[name] = v
        model.load_state_dict(new_state_dict)


# differentiate between the rainy picture and the normal picture by subtracting
# images from each other and finding different pixels and then
# deciding to allocate 0 rgb value to that pixel of 255
def diff(img1, img2):
    s = np.shape(img1)

    img1 = np.asarray(img1)
    img2 = np.asarray(img2)

    img1 = img1 / 255.
    img2 = img2 / 255.

    diff = 2 * (img1 - img2) + .5
    diff = diff.flatten()
    diff = [i if i < 1 else 1 for i in diff]
    diff = [i if i > 0 else 0 for i in diff]
    diff = np.reshape(diff, s)
    diff = diff * 255.
    return Image.fromarray(np.uint8(diff))


# The function for calculating PSNR(Peak signal-to-noise ratio) index
# which is used for the ratio between the maximum possible power
# of a signal and the power of corrupting noise
# that affects the fidelity of its representation.
def PSNR(original, compressed):
    original = np.asarray(original)
    compressed = np.asarray(compressed)

    mse = np.mean((original - compressed) ** 2)  # the formula for MSE(Mean squared error) index
    if mse == 0:
        return 100
    max_pixel = 255.0
    psnr = 20 * log10(max_pixel / sqrt(mse))
    return psnr


# The function for calculating RMSE(Root-mean-square deviation) index
# which is used to measure the differences between values
def RMSE(actual, predicted):
    actual = np.asarray(actual).flatten()
    predicted = np.asarray(predicted).flatten()

    MSE = np.square(np.subtract(actual, predicted)).mean()

    RMSE = sqrt(MSE)
    return RMSE


# We wrapped the main running part of the programm in try catch block, to dealing with any possible exceptions
try:
    input = sys.argv[1]  # input image
    output = sys.argv[2]    # output directory, which is used to save results

    t = time.time()
    current_path = os.path.dirname(os.path.abspath(__file__))
    weights = os.path.join(current_path, "models/Deraining.pth")  # loading weights

    load_file = run_path(os.path.join(current_path, "MPRNet/MPRNet_deraining.py"))
    model = load_file['MPRNet']()  # loading model

    load_checkpoint(model, weights)
    model.eval()  # model evaluation log

    img_multiple_of = 8

    img = Image.open(input).convert('RGB')
    # input_ = TF.to_tensor(img).unsqueeze(0).cuda() # for running with gpu
    input_ = TF.to_tensor(img).unsqueeze(0)

    # Pad the input if not_multiple_of 8
    h, w = input_.shape[2], input_.shape[3]
    H, W = ((h + img_multiple_of) // img_multiple_of) * img_multiple_of, (
            (w + img_multiple_of) // img_multiple_of) * img_multiple_of
    padh = H - h if h % img_multiple_of != 0 else 0
    padw = W - w if w % img_multiple_of != 0 else 0
    input_ = F.pad(input_, (0, padw, 0, padh), 'reflect')

    with torch.no_grad():
        restored = model(input_)
    restored = restored[0]
    restored = torch.clamp(restored, 0, 1)

    # Unpad the output
    restored = restored[:, :, :h, :w]

    restored = restored.permute(0, 2, 3, 1).cpu().detach().numpy()
    restored = img_as_ubyte(restored[0])

    # saving restored image
    cv2.imwrite(os.path.join(output, "result.png"), cv2.cvtColor(restored, cv2.COLOR_RGB2BGR))

    # calculate and save difference image
    img_diff = diff(restored, img)
    img_diff.save(os.path.join(output, 'diff.png'))

    # calculate and save psnr and rsme parameters
    psnr_restored = PSNR(img, restored)
    rsme_restored = RMSE(img, restored)

    f = open(os.path.join(output, "stdout.txt"), 'w')
    f.write("RMSE {:.2f}".format(rsme_restored))
    f.write("\nPSNR {:.2f}dB".format(psnr_restored))
    f.write("\nRuntime {:.2f}s".format(time.time() - t))
    f.close()

except:
    print("An error occurred, please check the input arguments.")
